/**
* jQuery asScroll v0.1.1
* https://github.com/amazingSurge/jquery-asScroll
*
* Copyright (c) amazingSurge
* Released under the LGPL-3.0 license
*/
import $ from 'jquery';

if (!Date.now) {
  Date.now = () => {
    return new Date().getTime();
  };
}

const vendors = ['webkit', 'moz'];
for (let i = 0; i < vendors.length && !window.requestAnimationFrame; ++i) {
  let vp = vendors[i];
  window.requestAnimationFrame = window[`${vp}RequestAnimationFrame`];
  window.cancelAnimationFrame = (window[`${vp}CancelAnimationFrame`] || window[`${vp}CancelRequestAnimationFrame`]);
}

if (/iP(ad|hone|od).*OS (6|7|8)/.test(window.navigator.userAgent) || !window.requestAnimationFrame || !window.cancelAnimationFrame) {
  let lastTime = 0;
  window.requestAnimationFrame = (callback) => {
    let now = getTime();
    let timePlus = 16;
    let nextTime = Math.max(lastTime + timePlus, now);
    return setTimeout(() => {
        callback(lastTime = nextTime);
      },
      nextTime - now);
  };
  window.cancelAnimationFrame = clearTimeout;
}

/**
 * Helper functions
 **/
function isPercentage (n) {
  'use strict';

  return typeof n === 'string' && n.indexOf('%') !== -1;
}

function convertPercentageToFloat(n) {
  'use strict';

  return parseFloat(n.slice(0, -1) / 100, 10);
}

function getTime() {
  if (typeof window.performance !== 'undefined' && window.performance.now) {
    return window.performance.now();
  }
  return Date.now();
}

var DEFAULTS = {
  duration: 800,
  easing: 'ease',
  namespace: 'asScroll',
  offsetTop: 50,
  mobile: {
    width: 768,
    duration: 500,
    easing: 'ease',
  }
};

const easingBezier = (mX1, mY1, mX2, mY2) => {
  'use strict';

  const a = (aA1, aA2) => {
    return 1.0 - 3.0 * aA2 + 3.0 * aA1;
  };

  const b = (aA1, aA2) => {
    return 3.0 * aA2 - 6.0 * aA1;
  };

  const c = (aA1) => {
    return 3.0 * aA1;
  };

  // Returns x(t) given t, x1, and x2, or y(t) given t, y1, and y2.
  const calcBezier = (aT, aA1, aA2) => {
    return ((a(aA1, aA2) * aT + b(aA1, aA2)) * aT + c(aA1)) * aT;
  };

  // Returns dx/dt given t, x1, and x2, or dy/dt given t, y1, and y2.
  const getSlope = (aT, aA1, aA2) => {
    return 3.0 * a(aA1, aA2) * aT * aT + 2.0 * b(aA1, aA2) * aT + c(aA1);
  };

  const getTForX = (aX) => {
    // Newton raphson iteration
    let aGuessT = aX;
    for (let i = 0; i < 4; ++i) {
      let currentSlope = getSlope(aGuessT, mX1, mX2);
      if (currentSlope === 0.0) {
        return aGuessT;
      }
      let currentX = calcBezier(aGuessT, mX1, mX2) - aX;
      aGuessT -= currentX / currentSlope;
    }
    return aGuessT;
  };

  if (mX1 === mY1 && mX2 === mY2) {
    return {
      css: 'linear',
      fn(aX) {
        return aX;
      }
    };
  }

  return {
    css: `cubic-bezier(${mX1},${mY1},${mX2},${mY2})`,
    fn(aX) {
      return calcBezier(getTForX(aX), mY1, mY2);
    }
  };
};

var EASING = {
  ease: easingBezier(0.25, 0.1, 0.25, 1.0),
  linear: easingBezier(0.00, 0.0, 1.00, 1.0),
  'ease-in': easingBezier(0.42, 0.0, 1.00, 1.0),
  'ease-out': easingBezier(0.00, 0.0, 0.58, 1.0),
  'ease-in-out': easingBezier(0.42, 0.0, 0.58, 1.0)
};

const NAMESPACE$1 = 'AsScroll';

let instanceId = 0;

const requestAnimFrame = (() => {
  return window.requestAnimationFrame ||
    window.webkitRequestAnimationFrame ||
    window.mozRequestAnimationFrame ||
    function(callback) {
        window.setTimeout(callback, 1000 / 60);
    };
})();

/**
 * Plugin constructor
 **/
class AsScroll {
  constructor(element, options) {
    this.element = element;
    this.$element = $(element);
    this.options = $.extend(true, {}, DEFAULTS, options);


    if (this.options.containerSelector) {
      this.$container = $(this.options.containerSelector);
    } else {
      this.$container = this.$element.is(document.body) ? $(window) : this.$element.parent();
    }
    if (this.$container.length !== 1) {
      return;
    }

    this.namespace = this.options.namespace;
    this.attributes = {
      vertical: {
        axis: 'Y',
        overflow: 'overflow-y',

        scroll: 'scrollTop',
        scrollLength: 'scrollHeight',
        pageOffset: 'pageYOffset',

        ffPadding: 'padding-right',

        length: 'height',
        clientLength: 'clientHeight',
        offsetLength: 'offsetHeight',
        offset: 'top',

        crossOffset: 'left',
        crossLength: 'width',
        crossClientLength: 'clientWidth',
        crossOffsetLength: 'offsetWidth'
      },
      horizontal: {
        axis: 'X',
        overflow: 'overflow-x',

        scroll: 'scrollLeft',
        scrollLength: 'scrollWidth',
        pageOffset: 'pageXOffset',

        ffPadding: 'padding-bottom',

        length: 'width',
        clientLength: 'clientWidth',
        offsetLength: 'offsetWidth',
        offset: 'left',

        crossOffset: 'top',
        crossLength: 'height',
        crossClientLength: 'clientHeight',
        crossOffsetLength: 'offsetHeight'
      }
    };

    this.classes = {};
    this.easing = EASING[this.options.easing] || EASING.ease;
    this.duration = this.options.duration;

    this._frameId = null;
    this._states = {};
    this.instanceId = (++instanceId);

    this.vertical = false;
    this.horizontal = false;

    this.init();
  }

  init() {
    const that = this;
    this.$targets = this.$element.find('[data-scroll-target]');
    this.lists = {};

    if (this.getScrollLength('vertical') > 0) {
      this.vertical = true;
    }

    if (this.getScrollLength('horizontal') > 0) {
      this.horizontal = true;
    }

    this.$targets.each(function() {
      const $target = $(this);

      that.lists[$target.data('scrollTarget')] = {
        $el: $target,
        offset: $target.offset()
      };
    });

    this.bindEvents();
  }

  getActiveTarget(direction) {
    /* eslint consistent-return: "off" */
    if (!this[direction]) {
      return;
    }

    const offset = this.getOffset(direction);
    const attributes = this.attributes[direction];
    let activeTarget = null;
    const containerLength = this.getContainerLength(direction);

    $.each(this.lists, (target, obj) => {
      const targetOffset = obj.offset[attributes.offset];
      const length = obj.$el[attributes.length]();

      if (targetOffset === offset) {
        activeTarget = target;
        return false;
      } else if (targetOffset > offset && targetOffset < offset + containerLength) {
        activeTarget = target;
        return false;
      } else if (targetOffset < offset && targetOffset + length > offset) {
        activeTarget = target;
        return true;
      } else if (targetOffset > offset + containerLength) {
        return false;
      }
    });

    return activeTarget;
  }

  bindEvents() {
    const that = this;

    $(window).on(this.eventNameWithId('resize'), () => {
      if ($(window).width() < that.options.mobile.width) {
        that.duration = that.options.duration;
        that.easing = AsScroll.easing[that.options.easing] || AsScroll.easing.ease;
      }
    });

    $(window).on(this.eventNameWithId('orientationchange'), () => {
      if ($(window).width() < that.options.mobile.width) {
        that.duration = that.options.duration;
        that.easing = AsScroll.easing[that.options.easing] || AsScroll.easing.ease;
      }
    });

    if (!this.horizontal && !this.vertical) {
      return;
    }

    this.$container.on(this.eventName('scroll'), () => {
      if (that.vertical) {
        that.trigger(that.eventName('active'), ['vertical', that.getActiveTarget('vertical')]);
      }
      if (that.horizontal) {
        that.getActiveTarget('horizontal')
        that.trigger(that.eventName('active'), ['horizontal', that.getActiveTarget('horizontal')]);
      }
    });
  }

  eventName(events) {
    if (typeof events !== 'string' || events === '') {
      return `.${this.options.namespace}`;
    }

    events = events.split(' ');
    const length = events.length;
    for (let i = 0; i < length; i++) {
      events[i] = `${events[i]}.${this.options.namespace}`;
    }
    return events.join(' ');
  }

  eventNameWithId(events) {
    if (typeof events !== 'string' || events === '') {
      return `${this.options.namespace}-${this.instanceId}`;
    }

    events = events.split(' ');
    const length = events.length;
    for (let i = 0; i < length; i++) {
      events[i] = `${events[i]}.${this.options.namespace}-${this.instanceId}`;
    }
    return events.join(' ');
  }

  trigger(eventType, ...params) {
    let data = [this].concat(...params);

    // event
    this.$element.trigger(`${NAMESPACE$1}::${eventType}`, data);

    // callback
    eventType = eventType.replace(/\b\w+\b/g, (word) => {
      return word.substring(0, 1).toUpperCase() + word.substring(1);
    });
    let onFunction = `on${eventType}`;

    if (typeof this.options[onFunction] === 'function') {
      this.options[onFunction].apply(this, ...params);
    }
  }

  /**
   * Checks whether the carousel is in a specific state or not.
   */
  is(state) {
    return this._states[state] && this._states[state] > 0;
  }

  /**
   * Enters a state.
   */
  enter(state) {
    if (this._states[state] === undefined) {
      this._states[state] = 0;
    }

    this._states[state]++;
  }

  /**
   * Leaves a state.
   */
  leave(state) {
    this._states[state]--;
  }

  getOffset(direction) {
    const attributes = this.attributes[direction];
    const element = this.$element[0];

    return (element[attributes.pageOffset] || element[attributes.scroll]);
  }

  getPercentOffset(direction) {
    return this.getOffset(direction) / this.getScrollLength(direction);
  }

  getContainerLength(direction) {
    return this.$container[0] === window ? this.$container[this.attributes[direction].length]() : this.$container[0][this.attributes[direction].clientLength];
  }

  getScrollLength(direction) {
    const scrollLength = this.$element[0][this.attributes[direction].scrollLength];
    return scrollLength - this.getContainerLength(direction);
  }

  scrollToTarget(direction, target, trigger, sync) {
    if (typeof this.lists[target] === 'undefined') {
      return;
    }
    target = this.lists[target];

    const attributes = this.attributes[direction];
    const offset = target.offset[attributes.offset];

    this.scrollTo(direction, offset, trigger, sync);
  }

  scrollTo(direction, value, trigger, sync) {
    let type = typeof value;

    if (type === "string") {
      if (isPercentage(value)) {
        value = convertPercentageToFloat(value) * this.getScrollLength(direction);
      }

      value = parseFloat(value);
      type = "number";
    }

    if (type !== "number") {
      return;
    }
    this.move(direction, value, trigger, sync);
  }

  scrollBy(direction, value, trigger, sync) {
    let type = typeof value;

    if (type === "string") {
      if (isPercentage(value)) {
        value = convertPercentageToFloat(value) * this.getScrollLength(direction);
      }

      value = parseFloat(value);
      type = "number";
    }

    if (type !== "number") {
      return;
    }

    this.move(direction, this.getOffset(direction) + value, trigger, sync);
  }

  move(direction, value, trigger, sync) {
    if (!this[direction] || typeof value !== "number") {
      return;
    }

    const that = this;

    this.enter('moving');

    if (value < 0) {
      value = 0;
    } else if (value > this.getScrollLength(direction)) {
      value = this.getScrollLength(direction);
    }

    const attributes = this.attributes[direction];

    const callback = () => {
      that.leave('moving');
    };

    if (sync) {

      this.$element[0][attributes.scroll] = value;

      if (trigger !== false) {
        this.trigger('change', value / this.getScrollLength(direction));
      }
      callback();
    } else {
      that.enter('animating');
      const startTime = getTime();
      const start = that.getOffset(direction);
      const end = value;

      const run = time => {
        let percent = (time - startTime) / that.duration;


        if (percent > 1) {
          percent = 1;
        }

        percent = that.easing.fn(percent);

        const current = parseFloat(start + percent * (end - start), 10);

        that.$element[0][attributes.scroll] = current;

        if (trigger !== false) {
          that.trigger('change', value / that.getScrollLength(direction));
        }

        if (percent === 1) {
          cancelAnimationFrame(that._frameId);
          that._frameId = null;

          that.leave('animating');
          callback();
        } else {
          that._frameId = requestAnimFrame(run);
        }
      };

      that._frameId = requestAnimFrame(run);
    }
  }

  scrollXto(value, trigger, sync) {
    return this.scrollTo('horizontal', value, trigger, sync);
  }

  scrollYto(value, trigger, sync) {
    return this.scrollTo('vertical', value, trigger, sync);
  }

  scrollXby(value, trigger, sync) {
    return this.scrollBy('horizontal', value, trigger, sync);
  }

  scrollYby(value, trigger, sync) {
    return this.scrollBy('vertical', value, trigger, sync);
  }

  scrollXToTarget(target, trigger, sync) {
    return this.scrollToTarget('horizontal', target, trigger, sync);
  }

  scrollYToTarget(target, trigger, sync) {
    return this.scrollToTarget('vertical', target, trigger, sync);
  }

  static registerEasing(name, ...args) {
    EASING[name] = easingBezier(...args);
  }

  static getEasing(name) {
    return EASING[name];
  }

  static setDefaults(options) {
    $.extend(true, DEFAULTS, $.isPlainObject(options) && options);
  }
}

var info = {
  version:'0.1.1'
};

const NAMESPACE = 'asScroll';
const OtherAsScroll = $.fn.asScroll;

const jQueryAsScroll = function(options, ...args) {
  if (typeof options === 'string') {
    const method = options;

    if (/^_/.test(method)) {
      return false;
    } else if ((/^(get)/.test(method))) {
      const instance = this.first().data(NAMESPACE);
      if (instance && typeof instance[method] === 'function') {
        return instance[method](...args);
      }
    } else {
      return this.each(function() {
        const instance = $.data(this, NAMESPACE);
        if (instance && typeof instance[method] === 'function') {
          instance[method](...args);
        }
      });
    }
  }

  return this.each(function() {
    if (!$(this).data(NAMESPACE)) {
      $(this).data(NAMESPACE, new AsScroll(this, options));
    }
  });
};

$.fn.asScroll = jQueryAsScroll;

$.asScroll = $.extend({
  setDefaults: AsScroll.setDefaults,
  registerEasing: AsScroll.registerEasing,
  getEasing: AsScroll.getEasing,
  noConflict: function() {
    $.fn.asScroll = OtherAsScroll;
    return jQueryAsScroll;
  }
}, info);
