import $ from 'jquery';
import * as util from './util';
import DEFAULTS from './defaults';
import easingBezier from './easingBezier';
import EASING from './easing';

const NAMESPACE = 'AsScroll';

let instanceId = 0;

const requestAnimFrame = (() => {
  return window.requestAnimationFrame ||
    window.webkitRequestAnimationFrame ||
    window.mozRequestAnimationFrame ||
    function(callback) {
        window.setTimeout(callback, 1000 / 60);
    };
})();

/**
 * Plugin constructor
 **/
class AsScroll {
  constructor(element, options) {
    this.element = element;
    this.$element = $(element);
    this.options = $.extend(true, {}, DEFAULTS, options);


    if (this.options.containerSelector) {
      this.$container = $(this.options.containerSelector);
    } else {
      this.$container = this.$element.is(document.body) ? $(window) : this.$element.parent();
    }
    if (this.$container.length !== 1) {
      return;
    }

    this.namespace = this.options.namespace;
    this.attributes = {
      vertical: {
        axis: 'Y',
        overflow: 'overflow-y',

        scroll: 'scrollTop',
        scrollLength: 'scrollHeight',
        pageOffset: 'pageYOffset',

        ffPadding: 'padding-right',

        length: 'height',
        clientLength: 'clientHeight',
        offsetLength: 'offsetHeight',
        offset: 'top',

        crossOffset: 'left',
        crossLength: 'width',
        crossClientLength: 'clientWidth',
        crossOffsetLength: 'offsetWidth'
      },
      horizontal: {
        axis: 'X',
        overflow: 'overflow-x',

        scroll: 'scrollLeft',
        scrollLength: 'scrollWidth',
        pageOffset: 'pageXOffset',

        ffPadding: 'padding-bottom',

        length: 'width',
        clientLength: 'clientWidth',
        offsetLength: 'offsetWidth',
        offset: 'left',

        crossOffset: 'top',
        crossLength: 'height',
        crossClientLength: 'clientHeight',
        crossOffsetLength: 'offsetHeight'
      }
    };

    this.classes = {};
    this.easing = EASING[this.options.easing] || EASING.ease;
    this.duration = this.options.duration;

    this._frameId = null;
    this._states = {};
    this.instanceId = (++instanceId);

    this.vertical = false;
    this.horizontal = false;

    this.init();
  }

  init() {
    const that = this;
    this.$targets = this.$element.find('[data-scroll-target]');
    this.lists = {};

    if (this.getScrollLength('vertical') > 0) {
      this.vertical = true;
    }

    if (this.getScrollLength('horizontal') > 0) {
      this.horizontal = true;
    }

    this.$targets.each(function() {
      const $target = $(this);

      that.lists[$target.data('scrollTarget')] = {
        $el: $target,
        offset: $target.offset()
      };
    });

    this.bindEvents();
  }

  getActiveTarget(direction) {
    /* eslint consistent-return: "off" */
    if (!this[direction]) {
      return;
    }

    const offset = this.getOffset(direction);
    const attributes = this.attributes[direction];
    let activeTarget = null;
    const containerLength = this.getContainerLength(direction);

    $.each(this.lists, (target, obj) => {
      const targetOffset = obj.offset[attributes.offset];
      const length = obj.$el[attributes.length]();

      if (targetOffset === offset) {
        activeTarget = target;
        return false;
      } else if (targetOffset > offset && targetOffset < offset + containerLength) {
        activeTarget = target;
        return false;
      } else if (targetOffset < offset && targetOffset + length > offset) {
        activeTarget = target;
        return true;
      } else if (targetOffset > offset + containerLength) {
        return false;
      }
    });

    return activeTarget;
  }

  bindEvents() {
    const that = this;

    $(window).on(this.eventNameWithId('resize'), () => {
      if ($(window).width() < that.options.mobile.width) {
        that.duration = that.options.duration;
        that.easing = AsScroll.easing[that.options.easing] || AsScroll.easing.ease;
      }
    });

    $(window).on(this.eventNameWithId('orientationchange'), () => {
      if ($(window).width() < that.options.mobile.width) {
        that.duration = that.options.duration;
        that.easing = AsScroll.easing[that.options.easing] || AsScroll.easing.ease;
      }
    });

    if (!this.horizontal && !this.vertical) {
      return;
    }

    this.$container.on(this.eventName('scroll'), () => {
      if (that.vertical) {
        that.trigger(that.eventName('active'), ['vertical', that.getActiveTarget('vertical')]);
      }
      if (that.horizontal) {
        that.getActiveTarget('horizontal')
        that.trigger(that.eventName('active'), ['horizontal', that.getActiveTarget('horizontal')]);
      }
    });
  }

  eventName(events) {
    if (typeof events !== 'string' || events === '') {
      return `.${this.options.namespace}`;
    }

    events = events.split(' ');
    const length = events.length;
    for (let i = 0; i < length; i++) {
      events[i] = `${events[i]}.${this.options.namespace}`;
    }
    return events.join(' ');
  }

  eventNameWithId(events) {
    if (typeof events !== 'string' || events === '') {
      return `${this.options.namespace}-${this.instanceId}`;
    }

    events = events.split(' ');
    const length = events.length;
    for (let i = 0; i < length; i++) {
      events[i] = `${events[i]}.${this.options.namespace}-${this.instanceId}`;
    }
    return events.join(' ');
  }

  trigger(eventType, ...params) {
    let data = [this].concat(...params);

    // event
    this.$element.trigger(`${NAMESPACE}::${eventType}`, data);

    // callback
    eventType = eventType.replace(/\b\w+\b/g, (word) => {
      return word.substring(0, 1).toUpperCase() + word.substring(1);
    });
    let onFunction = `on${eventType}`;

    if (typeof this.options[onFunction] === 'function') {
      this.options[onFunction].apply(this, ...params);
    }
  }

  /**
   * Checks whether the carousel is in a specific state or not.
   */
  is(state) {
    return this._states[state] && this._states[state] > 0;
  }

  /**
   * Enters a state.
   */
  enter(state) {
    if (this._states[state] === undefined) {
      this._states[state] = 0;
    }

    this._states[state]++;
  }

  /**
   * Leaves a state.
   */
  leave(state) {
    this._states[state]--;
  }

  getOffset(direction) {
    const attributes = this.attributes[direction];
    const element = this.$element[0];

    return (element[attributes.pageOffset] || element[attributes.scroll]);
  }

  getPercentOffset(direction) {
    return this.getOffset(direction) / this.getScrollLength(direction);
  }

  getContainerLength(direction) {
    return this.$container[0] === window ? this.$container[this.attributes[direction].length]() : this.$container[0][this.attributes[direction].clientLength];
  }

  getScrollLength(direction) {
    const scrollLength = this.$element[0][this.attributes[direction].scrollLength];
    return scrollLength - this.getContainerLength(direction);
  }

  scrollToTarget(direction, target, trigger, sync) {
    if (typeof this.lists[target] === 'undefined') {
      return;
    }
    target = this.lists[target];

    const attributes = this.attributes[direction];
    const offset = target.offset[attributes.offset];

    this.scrollTo(direction, offset, trigger, sync);
  }

  scrollTo(direction, value, trigger, sync) {
    let type = typeof value;

    if (type === "string") {
      if (util.isPercentage(value)) {
        value = util.convertPercentageToFloat(value) * this.getScrollLength(direction);
      }

      value = parseFloat(value);
      type = "number";
    }

    if (type !== "number") {
      return;
    }
    this.move(direction, value, trigger, sync);
  }

  scrollBy(direction, value, trigger, sync) {
    let type = typeof value;

    if (type === "string") {
      if (util.isPercentage(value)) {
        value = util.convertPercentageToFloat(value) * this.getScrollLength(direction);
      }

      value = parseFloat(value);
      type = "number";
    }

    if (type !== "number") {
      return;
    }

    this.move(direction, this.getOffset(direction) + value, trigger, sync);
  }

  move(direction, value, trigger, sync) {
    if (!this[direction] || typeof value !== "number") {
      return;
    }

    const that = this;

    this.enter('moving');

    if (value < 0) {
      value = 0;
    } else if (value > this.getScrollLength(direction)) {
      value = this.getScrollLength(direction);
    }

    const attributes = this.attributes[direction];

    const callback = () => {
      that.leave('moving');
    };

    if (sync) {

      this.$element[0][attributes.scroll] = value;

      if (trigger !== false) {
        this.trigger('change', value / this.getScrollLength(direction));
      }
      callback();
    } else {
      that.enter('animating');
      const startTime = util.getTime();
      const start = that.getOffset(direction);
      const end = value;

      const run = time => {
        let percent = (time - startTime) / that.duration;


        if (percent > 1) {
          percent = 1;
        }

        percent = that.easing.fn(percent);

        const current = parseFloat(start + percent * (end - start), 10);

        that.$element[0][attributes.scroll] = current;

        if (trigger !== false) {
          that.trigger('change', value / that.getScrollLength(direction));
        }

        if (percent === 1) {
          cancelAnimationFrame(that._frameId);
          that._frameId = null;

          that.leave('animating');
          callback();
        } else {
          that._frameId = requestAnimFrame(run);
        }
      };

      that._frameId = requestAnimFrame(run);
    }
  }

  scrollXto(value, trigger, sync) {
    return this.scrollTo('horizontal', value, trigger, sync);
  }

  scrollYto(value, trigger, sync) {
    return this.scrollTo('vertical', value, trigger, sync);
  }

  scrollXby(value, trigger, sync) {
    return this.scrollBy('horizontal', value, trigger, sync);
  }

  scrollYby(value, trigger, sync) {
    return this.scrollBy('vertical', value, trigger, sync);
  }

  scrollXToTarget(target, trigger, sync) {
    return this.scrollToTarget('horizontal', target, trigger, sync);
  }

  scrollYToTarget(target, trigger, sync) {
    return this.scrollToTarget('vertical', target, trigger, sync);
  }

  static registerEasing(name, ...args) {
    EASING[name] = easingBezier(...args);
  }

  static getEasing(name) {
    return EASING[name];
  }

  static setDefaults(options) {
    $.extend(true, DEFAULTS, $.isPlainObject(options) && options);
  }
}

export default AsScroll;
